@@ -9,7 +9,7 @@ import uuid
 from datetime import datetime
 from dateutil.tz import tzlocal
 
-from .common import UPLOAD_CHUNK_SIZE, InvalidProject
+from .common import UPLOAD_CHUNK_SIZE, InvalidProject, ClientError
 from .utils import generate_checksum, move_file, int_version, find, do_sqlite_checkpoint
 
 
@@ -22,10 +22,7 @@ this_dir = os.path.dirname(os.path.realpath(__file__))
 try:
     from .deps import pygeodiff
 except ImportError:
-    try:
-        import pygeodiff
-    except ImportError:
-        os.environ['GEODIFF_ENABLED'] = 'False'
+    import pygeodiff
 
 
 class MerginProject:
@@ -38,33 +35,31 @@ class MerginProject:
         if not os.path.exists(self.dir):
             raise InvalidProject('Project directory does not exist')
 
-        # make sure we can load correct pygeodiff
-        if os.environ.get('GEODIFF_ENABLED', 'True').lower() == 'true':
-            try:
-                self.geodiff = pygeodiff.GeoDiff()
-            except pygeodiff.geodifflib.GeoDiffLibVersionError:
-                self.geodiff = None
-        else:
-            self.geodiff = None
-
         self.meta_dir = os.path.join(self.dir, '.mergin')
         if not os.path.exists(self.meta_dir):
             os.mkdir(self.meta_dir)
 
         self.setup_logging(directory)
 
+        # make sure we can load correct pygeodiff
+        try:
+            self.geodiff = pygeodiff.GeoDiff()
+        except pygeodiff.geodifflib.GeoDiffLibVersionError:
+            # this is a fatal error, we can't live without geodiff
+            self.log.error("Unable to load geodiff! (lib version error)")
+            raise ClientError("Unable to load geodiff library!")
+
         # redirect any geodiff output to our log file
-        if self.geodiff:
-            def _logger_callback(level, text_bytes):
-                text = text_bytes.decode()  # convert bytes to str
-                if level == pygeodiff.GeoDiff.LevelError:
-                    self.log.error("GEODIFF: " + text)
-                elif level == pygeodiff.GeoDiff.LevelWarning:
-                    self.log.warning("GEODIFF: " + text)
-                else:
-                    self.log.info("GEODIFF: " + text)
-            self.geodiff.set_logger_callback(_logger_callback)
-            self.geodiff.set_maximum_logger_level(pygeodiff.GeoDiff.LevelDebug)
+        def _logger_callback(level, text_bytes):
+            text = text_bytes.decode()  # convert bytes to str
+            if level == pygeodiff.GeoDiff.LevelError:
+                self.log.error("GEODIFF: " + text)
+            elif level == pygeodiff.GeoDiff.LevelWarning:
+                self.log.warning("GEODIFF: " + text)
+            else:
+                self.log.info("GEODIFF: " + text)
+        self.geodiff.set_logger_callback(_logger_callback)
+        self.geodiff.set_maximum_logger_level(pygeodiff.GeoDiff.LevelDebug)
 
     def setup_logging(self, logger_name):
         """Setup logging into project directory's .mergin/client-log.txt file."""
@@ -127,8 +122,6 @@ class MerginProject:
         :returns: if file is compatible with geodiff lib
         :rtype: bool
         """
-        if not self.geodiff:
-            return False
         diff_extensions = ['.gpkg', '.sqlite']
         f_extension = os.path.splitext(file)[1]
         return f_extension in diff_extensions
@@ -247,14 +240,13 @@ class MerginProject:
         :returns: changes metadata for files to be pulled from server
         :rtype: dict
         """
-        changes = self.compare_file_sets(self.metadata['files'], server_files)
-        if not self.geodiff:
-            self.log.warning("geodiff is not available!")
-            return changes
 
+        # first let's have a look at the added/updated/removed files
+        changes = self.compare_file_sets(self.metadata['files'], server_files)
+
+        # then let's inspect our versioned files (geopackages) if there are any relevant changes
         not_updated = []
         for file in changes['updated']:
-            # for small geodiff files it does not make sense to download diff and then apply it (slow)
             if not self.is_versioned_file(file["path"]):
                 continue
 
@@ -310,10 +302,6 @@ class MerginProject:
                 file["checksum"] = checksum
             file['chunks'] = [str(uuid.uuid4()) for i in range(math.ceil(file["size"] / UPLOAD_CHUNK_SIZE))]
 
-        if not self.geodiff:
-            self.log.warning("geodiff is not available!")
-            return changes
-
         # need to check for for real changes in geodiff files using geodiff tool (comparing checksum is not enough)
         not_updated = []
         for file in changes['updated']:
@@ -344,12 +332,29 @@ class MerginProject:
                     not_updated.append(file)
             except (pygeodiff.GeoDiffLibError, pygeodiff.GeoDiffLibConflictError) as e:
                 self.log.warning("failed to create changeset for " + path)
-                # changes from wal file already committed
+                # probably the database schema has been modified if geodiff cannot create changeset.
+                # we will need to do full upload of the file
                 pass
 
         changes['updated'] = [f for f in changes['updated'] if f not in not_updated]
         return changes
 
+    def copy_versioned_file_for_upload(self, f, tmp_dir):
+        """
+        Make a temporary copy of the versioned file using geodiff, to make sure that we have full
+        content in a single file (nothing left in WAL journal)
+        """
+        path = f["path"]
+        self.log.info("Making a temporary copy (full upload): " + path)
+        tmp_file = os.path.join(tmp_dir, path)
+        os.makedirs(tmp_file, exist_ok=True)
+        self.geodiff.make_copy_sqlite(self.fpath(path), tmp_file)
+        f["size"] = os.path.getsize(tmp_file)
+        f["checksum"] = generate_checksum(tmp_file)
+        f["chunks"] = [str(uuid.uuid4()) for i in range(math.ceil(f["size"] / UPLOAD_CHUNK_SIZE))]
+        f["upload_file"] = tmp_file
+        return tmp_file
+
     def get_list_of_push_changes(self, push_changes):
         changes = {}
         for idx, file in enumerate(push_changes["updated"]):
@@ -374,7 +379,7 @@ class MerginProject:
         Update project files according to file changes. Apply changes to geodiff basefiles as well
         so they are up to date with server. In case of conflicts create backups from locally modified versions.
 
-        .. seealso:: self.pull_changes
+        .. seealso:: self.get_pull_changes
 
         :param changes: metadata for pulled files
         :type changes: dict[str, list[dict]]
@@ -410,18 +415,17 @@ class MerginProject:
 
                         # temporary backup of file pulled from server for recovery
                         f_server_backup = self.fpath(f'{path}-server_backup', temp_dir)
-                        shutil.copy(src, f_server_backup)
+                        self.geodiff.make_copy_sqlite(src, f_server_backup)
 
                         # create temp backup (ideally with geodiff) of locally modified file if needed later
                         f_conflict_file = self.fpath(f'{path}-local_backup', temp_dir)
                         try:
                             self.geodiff.create_changeset(basefile, dest, local_diff)
-                            shutil.copy(basefile, f_conflict_file)
+                            self.geodiff.make_copy_sqlite(basefile, f_conflict_file)
                             self.geodiff.apply_changeset(f_conflict_file, local_diff)
                         except (pygeodiff.GeoDiffLibError, pygeodiff.GeoDiffLibConflictError):
-                            self.log.warning("backup of local file with geodiff failed - need to do hard copy (dangerous!)")
-                            # FIXME hard copy can lead to data loss if changes from -wal file were not flushed !!!
-                            shutil.copy(dest, f_conflict_file)
+                            self.log.info("backup of local file with geodiff failed - need to do hard copy")
+                            self.geodiff.make_copy_sqlite(dest, f_conflict_file)
 
                         # in case there will be any conflicting operations found during rebase,
                         # they will be stored in a JSON file - if there are no conflicts, the file
@@ -438,17 +442,12 @@ class MerginProject:
                         except (pygeodiff.GeoDiffLibError, pygeodiff.GeoDiffLibConflictError) as err:
                             self.log.warning("rebase failed! going to create conflict file")
                             # it would not be possible to commit local changes, they need to end up in new conflict file
-                            shutil.copy(f_conflict_file, dest)  # revert file
+                            self.geodiff.make_copy_sqlite(f_conflict_file, dest)
                             conflict = self.backup_file(path)
                             conflicts.append(conflict)
                             # original file synced with server
-                            shutil.copy(f_server_backup, basefile)
-                            shutil.copy(f_server_backup, dest)
-                            # changes in -wal have been already applied in conflict file or LOST (see above)
-                            if os.path.exists(f'{dest}-wal'):
-                                os.remove(f'{dest}-wal')
-                            if os.path.exists(f'{dest}-shm'):
-                                os.remove(f'{dest}-shm')
+                            self.geodiff.make_copy_sqlite(f_server_backup, basefile)
+                            self.geodiff.make_copy_sqlite(f_server_backup, dest)
                     else:
                         # The local file is not modified -> no rebase needed.
                         # We just apply the diff between our copy and server to both the local copy and its basefile
@@ -464,12 +463,11 @@ class MerginProject:
                             self.geodiff.apply_changeset(basefile, server_diff)
                             self.log.info("update successful")
                         except (pygeodiff.GeoDiffLibError, pygeodiff.GeoDiffLibConflictError):
-                            self.log.warning("update failed! going to copy file (dangerous!)")
+                            self.log.warning("update failed! going to copy file")
                             # something bad happened and we have failed to patch our local files - this should not happen if there
                             # wasn't a schema change or something similar that geodiff can't handle.
-                            # FIXME: this is a last resort and may corrupt data! (we should warn user)
-                            shutil.copy(src, dest)
-                            shutil.copy(src, basefile)
+                            self.geodiff.make_copy_sqlite(src, dest)
+                            self.geodiff.make_copy_sqlite(src, basefile)
                 else:
                     # backup if needed
                     if path in modified and item['checksum'] != local_files_map[path]['checksum']:
@@ -485,9 +483,11 @@ class MerginProject:
                         if self.is_versioned_file(path):
                             os.remove(basefile)
                     else:
-                        shutil.copy(src, dest)
                         if self.is_versioned_file(path):
-                            shutil.copy(src, basefile)
+                            self.geodiff.make_copy_sqlite(src, dest)
+                            self.geodiff.make_copy_sqlite(src, basefile)
+                        else:
+                            shutil.copy(src, dest)
 
         return conflicts
 
@@ -498,8 +498,6 @@ class MerginProject:
         :param changes: metadata for pulled files
         :type changes: dict[str, list[dict]]
         """
-        if not self.geodiff:
-            return
         for k, v in changes.items():
             for item in v:
                 path = item['path']
@@ -510,12 +508,12 @@ class MerginProject:
                 if k == 'removed':
                     os.remove(basefile)
                 elif k == 'added':
-                    shutil.copy(self.fpath(path), basefile)
+                    self.geodiff.make_copy_sqlite(self.fpath(path), basefile)
                 elif k == 'updated':
                     # in case for geopackage cannot be created diff (e.g. forced update with committed changes from wal file)
                     if "diff" not in item:
                         self.log.info("updating basefile (copy) for: " + path)
-                        shutil.copy(self.fpath(path), basefile)
+                        self.geodiff.make_copy_sqlite(self.fpath(path), basefile)
                     else:
                         self.log.info("updating basefile (diff) for: " + path)
                         # better to apply diff to previous basefile to avoid issues with geodiff tmp files
@@ -545,7 +543,10 @@ class MerginProject:
         while os.path.exists(backup_path):
             backup_path = self.fpath(f'{file}_conflict_copy{index}')
             index += 1
-        shutil.copy(src, backup_path)
+        if self.is_versioned_file(file):
+            self.geodiff.make_copy_sqlite(src, backup_path)
+        else:
+            shutil.copy(src, backup_path)
         return backup_path
 
     def apply_diffs(self, basefile, diffs):
